import argparse
import json
from collections import OrderedDict, deque
from statistics import mean
from typing import Literal
from uuid import uuid4

import zmq.asyncio
from rich.table import Table
from textual import on, work
from textual.app import App, ComposeResult
from textual.binding import Binding
from textual.containers import Grid, Horizontal, Vertical, VerticalScroll
from textual.events import Resize
from textual.screen import ModalScreen
from textual.widgets import (
    Button,
    Collapsible,
    ContentSwitcher,
    DataTable,
    Digits,
    Header,
    Label,
    ProgressBar,
    Static,
    TabbedContent,
    TabPane,
    TextArea,
    Tree,
)
from textual.widgets.data_table import ColumnKey
from textual.worker import get_current_worker

from mrok import __version__
from mrok.agent.devtools.inspector.utils import (
    build_tree,
    get_highlighter_language_by_content_type,
    hexdump,
    humanize_bytes,
    parse_content_type,
    parse_form_data,
)
from mrok.proxy.models import (
    Event,
    HTTPHeaders,
    HTTPRequest,
    HTTPResponse,
    ServiceMetadata,
    WorkerMetrics,
)

MIN_COLS = 160
MIN_ROWS = 45


class Counter(Digits):
    DEFAULT_CSS = """
    Counter {
        text-align: center;
        border: round #00BBFF;
    }
    """

    def __init__(
        self,
        counter_name: str,
        value: str = "",
        *,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        self.counter_name = counter_name
        super().__init__(
            value,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

    def on_mount(self) -> None:
        self.border_title = self.counter_name


class ProcessInfoBar(Static):
    DEFAULT_CSS = """
    ProcessInfoBar {
        width: 100%;
        align: center middle;
        padding: 1;
    }
    ProgressBar {
        width: 100%;
    }
    ProgressBar > Bar {
        width: 24;
    }
    ProgressBar > Bar > .bar--complete {
        color: red;
    }
    Vertical {
        width: auto;
        align: center middle;
    }
    """

    def __init__(self, label: str, **kwargs):
        super().__init__(**kwargs)
        self.label_text = label

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Static(f"[b]{self.label_text}[/b]")
            yield ProgressBar(
                total=100,
                show_percentage=True,
                show_eta=False,
            )

    def update_value(self, value: float) -> None:
        self.query_one(ProgressBar).update(progress=value)


class InfoPanel(Static):
    BORDER_TITLE = "Info"

    DEFAULT_CSS = """
    InfoPanel {
        background: black;
        height: 100%;
        border: round #0088FF;
    }
    """

    def __init__(
        self,
        content="",
        *,
        expand=False,
        shrink=False,
        markup=True,
        name=None,
        id=None,
        classes=None,
        disabled=False,
    ) -> None:
        super().__init__(
            content,
            expand=expand,
            shrink=shrink,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )

    def compose(self) -> ComposeResult:
        yield DataTable(show_header=False, cursor_type="none")

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        table.add_columns("Label", "Value")

    def update_meta(self, meta: ServiceMetadata) -> None:
        table = self.query_one(DataTable)
        if len(table.rows) == 0:
            table.add_row("URL", f"https://{meta.extension}.{meta.domain}")
            table.add_row("Extension ID", meta.extension.upper())
            table.add_row("Instance ID", meta.instance.upper())
            table.add_row("mrok version", __version__)
            self.loading = False


class Requests(Static):
    BORDER_TITLE = "Requests"
    DEFAULT_CSS = """
    Requests {
        background: black;
        height: 100%;
        border: round #0088FF;
    }
    """

    def compose(self) -> ComposeResult:
        # give the table an id so CSS can target it; enable zebra stripes
        yield DataTable(id="requests", zebra_stripes=True, cursor_type="row")

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        table.add_column(
            "Method",
            width=7,
        )
        table.add_column(
            "Status",
            width=6,
        )
        table.add_column("Path", width=25, key="path")
        table.add_column(
            "Duration (ms)",
            width=13,
        )


class Details(Static):
    BORDER_TITLE = "Details"

    def compose(self) -> ComposeResult:
        with TabbedContent():
            with TabPane("Headers"):
                with VerticalScroll():
                    with Collapsible(title="General", collapsed=False):
                        yield DataTable(id="general-data", show_header=False, show_cursor=False)

                    with Collapsible(
                        title="Request Headers", collapsed=False, id="request-headers"
                    ):
                        yield Static(id="request-headers-table")
                    with Collapsible(
                        title="Response Headers", collapsed=False, id="response-headers"
                    ):
                        yield Static(id="response-headers-table")
            with TabPane("Payload"):
                with ContentSwitcher(initial="payload-tree", id="payload-switcher"):
                    yield Tree("Payload", id="payload-tree")
                    yield DataTable(show_header=False, show_cursor=False, id="payload-formdata")
                    yield TextArea(id="payload-other", read_only=True)
            with TabPane("Preview"):
                with ContentSwitcher(initial="preview-tree", id="preview-switcher"):
                    yield Tree("Preview", id="preview-tree")
                    yield TextArea(id="preview-other", read_only=True)
            with TabPane("Response"):
                yield TextArea(id="raw-response", read_only=True)

    def update_headers(self, type: Literal["request", "response"], headers: HTTPHeaders):
        collapsible = self.query_one(f"#{type}-headers", Collapsible)
        collapsible.title = (
            f"{type.capitalize()} Headers ({len(headers)})"
            if len(headers)
            else f"{type.capitalize()} Headers"
        )
        if len(headers):
            headers_table = Table(show_header=False, box=None)
            for k, v in headers.items():
                headers_table.add_row(f"[orange3]{k}[/]", v)
            self.query_one(f"#{type}-headers-table", Static).content = headers_table
        else:
            self.query_one(f"#{type}-headers-table", Static).content = ""

    def update_preview(self, response: HTTPResponse):
        tree = self.query_one("#preview-tree", Tree)
        tree.clear()
        self.query_one("#preview-switcher", ContentSwitcher).current = "preview-tree"
        if not response.body:
            return
        content_type_info = parse_content_type(response.headers["Content-Type"])
        if content_type_info.content_type == "application/json":
            build_tree(
                tree.root,
                json.loads(response.body.decode(encoding=content_type_info.charset)),  # type: ignore[arg-type]
            )
            return
        text = (
            hexdump(response.body)
            if content_type_info.binary
            else response.body.decode(encoding=content_type_info.charset)  # type: ignore[arg-type]
        )
        txtarea = self.query_one("#preview-other", TextArea)
        txtarea.text = text
        txtarea.language = get_highlighter_language_by_content_type(content_type_info.content_type)
        self.query_one("#preview-switcher", ContentSwitcher).current = "preview-other"

    def update_payload(self, request: HTTPRequest):
        tree = self.query_one("#payload-tree", Tree)
        tree.clear()
        form_data_table = self.query_one("#payload-formdata", DataTable)
        form_data_table.clear()
        self.query_one("#payload-switcher", ContentSwitcher).current = "payload-tree"
        if not request.body:
            return
        content_type_info = parse_content_type(request.headers["Content-Type"])
        if content_type_info.content_type == "application/json":
            build_tree(
                tree.root,
                json.loads(request.body.decode(encoding=content_type_info.charset)),  # type: ignore[arg-type]
            )
            return
        if content_type_info.content_type == "multipart/form-data" and not request.body_truncated:
            for name, value in parse_form_data(request.body, content_type_info.boundary):  # type: ignore[arg-type]
                form_data_table.add_row(name, value)
            self.query_one("#payload-switcher", ContentSwitcher).current = "payload-formdata"
            return
        text = (
            hexdump(request.body)
            if content_type_info.binary
            else request.body.decode(encoding=content_type_info.charset)  # type: ignore[arg-type]
        )
        txtarea = self.query_one("#payload-other", TextArea)
        txtarea.text = text
        self.query_one("#payload-switcher", ContentSwitcher).current = "payload-other"

    def update_response(self, response: HTTPResponse):
        text_area = self.query_one("#raw-response", TextArea)
        text_area.clear()
        if response.body:
            content_type_info = parse_content_type(response.headers["Content-Type"])
            text = (
                hexdump(response.body)
                if content_type_info.binary
                else response.body.decode(encoding=content_type_info.charset)  # type: ignore[arg-type]
            )
            text_area.text = text
            text_area.language = get_highlighter_language_by_content_type(
                content_type_info.content_type
            )

    def update_info(self, response: HTTPResponse):
        general_table = self.query_one("#general-data", DataTable)
        general_table.clear()
        url = response.request.url
        if response.request.query_string:
            url = f"{url}?{response.request.query_string.decode()}"
        general_table.add_row("Request Url", url)
        general_table.add_row("Request Method", self.format_method(response.request.method))
        general_table.add_row("Status Code", self.format_status(response.status))
        self.update_headers("request", response.request.headers)
        self.update_headers("response", response.headers)
        self.update_payload(response.request)
        self.update_preview(response)
        self.update_response(response)

    def format_status(self, status: int) -> str:
        if status < 400:
            return f"🟢 [green]{status}[/]"
        elif status < 500:
            return f"🟠 [orange3]{status}[/]"
        else:
            return f"🔴 [red]{status}[/]"

    def format_method(self, method: str) -> str:
        method = method.upper()
        if method in ["POST", "PUT", "PATCH"]:
            return f"[orange3]{method}[/]"
        elif method == "DELETE":
            return f"[red]{method}[/]"
        else:
            return f"[blue]{method}[/]"

    def on_mount(self) -> None:
        general_table = self.query_one("#general-data", DataTable)
        general_table.add_columns("Label", "Value")
        form_data_table = self.query_one("#payload-formdata", DataTable)
        form_data_table.add_columns("Name", "Value")


class LeftPanel(Static):
    DEFAULT_CSS = """
    LeftPanel {
        layout: grid;
        grid-size: 1 3;
        grid-rows: 6 2fr 4fr;
        grid-columns: 1fr;
        background: black;
        height: 100%;
        width: 5fr;
    }
    Details {
        background: black;
        height: 100%;
        border: round #0088FF;
    }
    """

    def compose(self):
        yield InfoPanel()
        yield Requests()
        yield Details()


class ProcessMetrics(Static):
    BORDER_TITLE = "Process"
    DEFAULT_CSS = """
    ProcessMetrics {
        border: round #00BBFF;
        layout: grid;
        grid-size: 1 2;
        grid-gutter: 0;
        height: 100%;
        align: center middle;
    }
    """

    def __init__(
        self,
        content="",
        *,
        expand=False,
        shrink=False,
        markup=True,
        name=None,
        id=None,
        classes=None,
        disabled=False,
    ) -> None:
        super().__init__(
            content,
            expand=expand,
            shrink=shrink,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.mem_values: deque = deque([0] * 10, maxlen=10)
        self.cpu_values: deque = deque([0] * 10, maxlen=10)
        self.workers_metrics: dict[str, WorkerMetrics] = {}

    def compose(self) -> ComposeResult:
        yield ProcessInfoBar("CPU", id="cpu")
        yield ProcessInfoBar("Memory", id="mem")

    def on_mount(self) -> None:
        self.query_one("#cpu", ProcessInfoBar).update_value(0)
        self.query_one("#mem", ProcessInfoBar).update_value(0)

    def update_stats(self) -> None:
        cpu = mean([m.process.cpu for m in self.workers_metrics.values()])
        mem = mean([m.process.mem for m in self.workers_metrics.values()])

        self.query_one("#cpu", ProcessInfoBar).update_value(cpu)
        self.query_one("#mem", ProcessInfoBar).update_value(mem)

    def update_metrics(self, metrics: WorkerMetrics) -> None:
        self.workers_metrics[metrics.worker_id] = metrics
        self.update_stats()


class RequestsMetrics(Static):
    BORDER_TITLE = "Requests"
    DEFAULT_CSS = """
    RequestsMetrics {
        background: black;
        layout: grid;
        grid-size: 1 4;
        grid-gutter: 0;
        align-vertical: middle;
        border: round #0088FF;
        height: 100%;
        grid-rows: auto auto auto auto;
    }
    #requests-rps {
        color: #0088FF;
    }
    #requests-ok {
        color: green;
    }
    #requests-ko {
        color: red;
    }
    """

    def __init__(
        self,
        content="",
        *,
        expand=False,
        shrink=False,
        markup=True,
        name=None,
        id=None,
        classes=None,
        disabled=False,
    ) -> None:
        super().__init__(
            content,
            expand=expand,
            shrink=shrink,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.workers_metrics: dict[str, WorkerMetrics] = {}

    def compose(self) -> ComposeResult:
        yield Counter("RPS", id="requests-rps")
        yield Counter("Total", id="requests-total")
        yield Counter("Success", id="requests-ok")
        yield Counter("Failed", id="requests-ko")

    def on_mount(self) -> None:
        self.update_stats()

    def update_stats(self) -> None:
        self.query_one("#requests-rps", Digits).update(
            str(sum(m.requests.rps for m in self.workers_metrics.values()))
        )
        self.query_one("#requests-total", Digits).update(
            str(sum(m.requests.total for m in self.workers_metrics.values()))
        )
        self.query_one("#requests-ok", Digits).update(
            str(sum(m.requests.successful for m in self.workers_metrics.values()))
        )
        self.query_one("#requests-ko", Digits).update(
            str(sum(m.requests.failed for m in self.workers_metrics.values()))
        )
        self.loading = False

    def update_metrics(self, metrics: WorkerMetrics) -> None:
        self.workers_metrics[metrics.worker_id] = metrics
        self.update_stats()


class DataTransferMetrics(Static):
    BORDER_TITLE = "Transfer"
    DEFAULT_CSS = """
    DataTransferMetrics {
        background: black;
        layout: grid;
        grid-size: 1 2;
        grid-gutter: 0;
        align-vertical: middle;
        border: round #0088FF;
        height: 100%;
        grid-rows: auto auto auto auto;
    }
    """

    def __init__(
        self,
        content="",
        *,
        expand=False,
        shrink=False,
        markup=True,
        name=None,
        id=None,
        classes=None,
        disabled=False,
    ) -> None:
        super().__init__(
            content,
            expand=expand,
            shrink=shrink,
            markup=markup,
            name=name,
            id=id,
            classes=classes,
            disabled=disabled,
        )
        self.workers_metrics: dict[str, WorkerMetrics] = {}

    def compose(self) -> ComposeResult:
        yield Counter("In", id="in")
        yield Counter("Out", id="out")

    def on_mount(self) -> None:
        self.update_stats()

    def update_stats(self) -> None:
        amount, unit = humanize_bytes(
            sum(m.data_transfer.bytes_in for m in self.workers_metrics.values())
        )
        self.query_one("#in", Digits).border_title = f"In ({unit})"
        self.query_one("#in", Digits).update(str(amount))
        amount, unit = humanize_bytes(
            sum(m.data_transfer.bytes_out for m in self.workers_metrics.values())
        )
        self.query_one("#out", Digits).border_title = f"Out ({unit})"
        self.query_one("#out", Digits).update(str(amount))
        self.loading = False

    def update_metrics(self, metrics: WorkerMetrics) -> None:
        self.workers_metrics[metrics.worker_id] = metrics
        self.update_stats()


class RightPanel(Static):
    DEFAULT_CSS = """
    RightPanel {
        layout: grid;
        grid-size: 1 3;
        grid-rows: 1fr 22 12;
        grid-columns: 1fr;
        background: black;
        height: 100%;
        width: 33;
    }
    .hidden {
        display: none;
    }
    """

    def compose(self):
        yield ProcessMetrics()
        yield RequestsMetrics()
        yield DataTransferMetrics()


class TooSmallScreen(ModalScreen):
    CSS = """
    TooSmallScreen {
        align: center middle;
    }
    #dialog {
        grid-size: 1 3;
        grid-rows: 1fr 2fr 3;
        padding: 0 1;
        width: 60;
        height: 15;
        border: thick $background 80%;
        background: $surface;
    }
    #title {
        column-span: 1;
        height: 1fr;
        width: 1fr;
        content-align: center middle;
        background: $panel;
        color: $foreground;
        text-style: bold;
    }
    #message {
        margin-top: 1;
        column-span: 1;
        height: 1fr;
        width: 1fr;
        content-align: center top;
    }
    Button {
        width: 100%;
    }
    """

    def __init__(self):
        super().__init__()
        self.dialog_title = "Terminal too small"
        self.dialog_message = (
            f"Your current terminal size is {self.app.size.width}x{self.app.size.height}. "
            f"For the best experience please resize your terminal to {MIN_COLS}x{MIN_ROWS}."
        )
        self.btn_label = "dismiss"
        self.btn_variant = "primary"

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(self.dialog_title, id="title"),
            Label(self.dialog_message, id="message"),
            Button(self.btn_label, variant=self.btn_variant, id="dismiss"),
            id="dialog",
        )

    @on(Button.Pressed)
    def on_button_pressed(self, event: Button.Pressed) -> None:
        event.stop()
        if event.button.id == "dismiss":
            self.app.pop_screen()

    def update_message(self):
        self.query_one("#message", Label).content = (
            f"Your current terminal size is {self.app.size.width}x{self.app.size.height}. "
            f"For the best experience please resize your terminal to {MIN_COLS}x{MIN_ROWS}."
        )


class InspectorApp(App):
    TITLE = "mrok Dev Console"
    CSS = """
    Screen {
        background: black;
    }
    # Screen {
    #     layout: grid;
    #     grid-size: 2 1;
    #     grid-columns: 5fr 1fr;
    #     background: black;
    # }
    DataTable {
        background: black;
        color: #dddddd;
        height: auto;
        max-height: 100%;
    }

    /* Header row: white title text, neon underline */
    DataTable > .datatable--header {
        text-style: bold;
        background: black;
        color: white;
        border-bottom: solid #0088FF;
    }

    /* Cells inherit neon text on black */
    DataTable .cell {
        color: #0088FF;
        background: black;
    }

    /* Zebra stripes: even rows slightly lighter than black */
    DataTable > .datatable--even-row {
        background: rgb(10,10,14);
    }
    DataTable > .datatable--odd-row {
        background: black;
    }

    /* Hover and focus styles */
    DataTable > .datatable--hover {
        background: rgb(6,12,20);
    }
    DataTable:focus > .datatable--cursor {
        background: rgb(2,40,70);
        color: white;
    }

    /* Fixed columns/rows use a muted neon background */
    DataTable > .datatable--fixed {
        background: rgb(6,6,10);
        color: #00CFFF;
    }
    Collapsible {
        background: black;
        border: round #00BBFF;
    }
    Tree {
        background: black;
    }
    TextArea {
        background: black;
    }
    RightPanel.-hidden {
        display: none;
    }
    """

    BINDINGS = [
        Binding("m", "toggle_metrics()", "Toggle Metrics"),
    ]

    def __init__(
        self,
        subscriber_port: int,
        max_requests: int = 1000,
        driver_class=None,
        css_path=None,
        watch_css=False,
        ansi_color=False,
    ):
        super().__init__(driver_class, css_path, watch_css, ansi_color)
        self.ctx = zmq.asyncio.Context()
        self.socket = self.ctx.socket(zmq.SUB)
        self.socket.connect(f"tcp://127.0.0.1:{subscriber_port}")
        self.max_requests = max_requests
        self.requests: OrderedDict = OrderedDict()

    def compose(self) -> ComposeResult:
        yield Header()
        with Horizontal():
            yield LeftPanel()
            yield RightPanel()

    async def on_mount(self):
        self._check_minimum_size(self.size.width, self.size.height)
        self.socket.subscribe("")
        self.consumer()

    async def on_unmount(self):
        self.socket.close()
        self.ctx.term()

    @work(exclusive=True)
    async def consumer(self):
        worker = get_current_worker()
        requests_table = self.query_one("#requests", DataTable)
        while not worker.is_cancelled:
            try:
                event = Event.model_validate_json(await self.socket.recv_string())
                if event.type == "status":
                    info_widget = self.query_one(InfoPanel)
                    info_widget.update_meta(event.data.meta)
                    process_metrics_widget = self.query_one(ProcessMetrics)
                    process_metrics_widget.update_metrics(event.data.metrics)
                    requests_metrics_widget = self.query_one(RequestsMetrics)
                    requests_metrics_widget.update_metrics(event.data.metrics)
                    data_transfer_metrics_widget = self.query_one(DataTransferMetrics)
                    data_transfer_metrics_widget.update_metrics(event.data.metrics)
                    continue

                response = event.data
                if len(self.requests) == self.max_requests:
                    self.requests.popitem(last=False)
                req_id = str(uuid4())
                self.requests[req_id] = response
                requests_table.clear()
                for key, response in reversed(self.requests.items()):
                    requests_table.add_row(
                        response.request.method,
                        response.status,
                        response.request.url,
                        response.duration * 1000,
                        key=key,
                    )
            except Exception as e:
                self.log(e)

    def action_toggle_metrics(self):
        self.query_one(RightPanel).toggle_class("-hidden")

    def on_data_table_row_selected(self, event: DataTable.RowSelected) -> None:
        if not event.row_key:
            return
        response = self.requests[event.row_key]
        details = self.query_one(Details)
        details.update_info(response)

    @on(Resize)
    def handle_resize(self, event: Resize):
        self._check_minimum_size(event.size.width, event.size.height)
        width = event.size.width
        table = self.query_one("#requests", DataTable)
        table.columns[ColumnKey("path")].width = width - 69

    def _check_minimum_size(self, width: int, height: int):
        if width < MIN_COLS or height < MIN_ROWS:
            if not isinstance(self.screen, TooSmallScreen):
                self.push_screen(TooSmallScreen())
            else:
                self.screen.update_message()
            return

        if isinstance(self.screen, TooSmallScreen):
            self.pop_screen()


def module_main(argv: list[str]) -> None:
    parser = argparse.ArgumentParser(description="mrok devtools agent")
    parser.add_argument(
        "-p",
        "--subscriber-port",
        type=int,
        default=50001,
        help="Port for subscriber (default: 50001)",
    )
    args = parser.parse_args(argv)
    app = InspectorApp(args.subscriber_port)
    app.run()



from collections.abc import Generator
from dataclasses import dataclass
from io import BytesIO

from multipart import MultipartParser

TEXTUAL_CONTENT_TYPES = {
    "application/json",
    "application/xml",
    "application/javascript",
    "application/x-www-form-urlencoded",
}

TEXTUAL_PREFIXES = ("text/",)

CONTENT_TYPE_TO_LANGUAGE = {
    "application/json": "json",
    "application/ld+json": "json",
    "application/problem+json": "json",
    "application/schema+json": "json",
    "application/xml": "xml",
    "text/xml": "xml",
    "application/xhtml+xml": "html",
    "text/html": "html",
    "text/css": "css",
    "application/javascript": "javascript",
    "application/x-javascript": "javascript",
    "text/javascript": "javascript",
    "application/ecmascript": "javascript",
    "text/markdown": "markdown",
    "text/x-markdown": "markdown",
    "application/yaml": "yaml",
    "application/x-yaml": "yaml",
    "text/yaml": "yaml",
    "application/toml": "toml",
    "application/x-toml": "toml",
    "application/sql": "sql",
    "text/x-sql": "sql",
    "application/java": "java",
    "text/x-java-source": "java",
    "application/python": "python",
    "text/x-python": "python",
    "application/x-python-code": "python",
    "application/rust": "rust",
    "text/x-rust": "rust",
    "application/go": "go",
    "text/x-go": "go",
    "application/bash": "bash",
    "application/x-sh": "bash",
    "text/x-shellscript": "bash",
    "application/regex": "regex",
    "text/x-regex": "regex",
}


@dataclass
class ContentTypeInfo:
    content_type: str
    binary: bool
    charset: str | None = None
    boundary: str | None = None


def parse_content_type(content_type_header: str) -> ContentTypeInfo:
    parts = content_type_header.split(";")
    content_type = parts[0].strip().lower()

    charset = None
    boundary = None

    for part in parts[1:]:
        part = part.strip()
        if "=" in part:
            key, value = part.split("=", 1)
            key = key.strip().lower()
            value = value.strip().strip('"')
            if key == "charset":
                charset = value
            elif key == "boundary":
                boundary = value

    binary = not is_textual(content_type)

    if charset is None and not binary:
        charset = "utf-8"

    return ContentTypeInfo(
        content_type=content_type, binary=binary, charset=charset, boundary=boundary
    )


def parse_form_data(data: bytes, boundary: str) -> Generator[tuple[str, str]]:
    parser = MultipartParser(BytesIO(data), boundary)
    for part in parser:
        if is_textual(part.content_type):
            yield part.name, part.value
            continue
        yield part.name, "<binary>"


def is_textual(content_type: str) -> bool:
    ct = content_type.lower()
    if ct in TEXTUAL_CONTENT_TYPES:
        return True
    if any(ct.startswith(p) for p in TEXTUAL_PREFIXES):
        return True
    return False


def build_tree(node, data):
    if isinstance(data, dict):
        for key, value in data.items():
            child = node.add(str(key))
            build_tree(child, value)
    elif isinstance(data, list):
        for index, value in enumerate(data):
            child = node.add(f"[{index}]")
            build_tree(child, value)
    else:
        node.add(repr(data))


def hexdump(data, width=16):
    lines = []
    for i in range(0, len(data), width):
        chunk = data[i : i + width]
        hex_part = " ".join(f"{b:02x}" for b in chunk)
        ascii_part = "".join(chr(b) if 32 <= b <= 126 else "." for b in chunk)
        lines.append(f"{hex_part:<{width * 3}} {ascii_part}")
    return "\n".join(lines)


def humanize_bytes(num_bytes: int) -> tuple[float, str]:  # type: ignore[return-value]
    if num_bytes < 0:
        raise ValueError("num_bytes must be non-negative")

    units = ["B", "KiB", "MiB", "GiB", "TiB", "PiB"]
    value = float(num_bytes)

    for unit in units:
        if value < 1024 or unit == units[-1]:
            return round(value, 2), unit
        value /= 1024


def get_highlighter_language_by_content_type(content_type: str) -> str | None:
    if content_type in CONTENT_TYPE_TO_LANGUAGE:
        return CONTENT_TYPE_TO_LANGUAGE[content_type]
    return None



import asyncio
import contextlib
import logging
from pathlib import Path

from uvicorn.importer import import_from_string

from mrok.conf import get_settings
from mrok.logging import setup_logging
from mrok.proxy.asgi import ASGIAppWrapper
from mrok.proxy.events import EventsPublisher
from mrok.proxy.models import Identity
from mrok.proxy.ziticorn import BackendConfig, Server
from mrok.types.proxy import ASGIApp

logger = logging.getLogger("mrok.proxy")


class Worker:
    def __init__(
        self,
        worker_id: str,
        app: ASGIApp | str,
        identity_file: str | Path,
        *,
        ziti_load_timeout_ms: int = 5000,
        server_backlog: int = 2048,
        server_timeout_keep_alive: int = 5,
        server_limit_concurrency: int | None = None,
        server_limit_max_requests: int | None = None,
        events_enabled: bool = True,
        events_publisher_port: int = 50000,
        events_metrics_collect_interval: float = 5.0,
    ):
        self._worker_id = worker_id
        self._identity_file = identity_file
        self._identity = Identity.load_from_file(self._identity_file)
        self._app = app
        self._ziti_load_timeout_ms = ziti_load_timeout_ms
        self._server_backlog = server_backlog
        self._server_timeout_keep_alive = server_timeout_keep_alive
        self._server_limit_concurrency = server_limit_concurrency
        self._server_limit_max_requests = server_limit_max_requests

        self._events_enabled = events_enabled
        self._event_publisher = (
            EventsPublisher(
                worker_id=worker_id,
                meta=self._identity.mrok,
                events_publisher_port=events_publisher_port,
                events_metrics_collect_interval=events_metrics_collect_interval,
            )
            if events_enabled
            else None
        )

    def setup_app(self):
        app = ASGIAppWrapper(
            self._app if not isinstance(self._app, str) else import_from_string(self._app),
            lifespan=self._event_publisher.lifespan if self._events_enabled else None,
        )

        if self._events_enabled:
            self._event_publisher.setup_middleware(app)
        return app

    def run(self):
        setup_logging(get_settings())
        app = self.setup_app()

        config = BackendConfig(
            app,
            self._identity_file,
            ziti_load_timeout_ms=self._ziti_load_timeout_ms,
            backlog=self._server_backlog,
            timeout_keep_alive=self._server_timeout_keep_alive,
            limit_concurrency=self._server_limit_concurrency,
            limit_max_requests=self._server_limit_max_requests,
        )
        server = Server(config)
        with contextlib.suppress(KeyboardInterrupt, asyncio.CancelledError):
            server.run()



import logging
import socket
from collections.abc import Callable
from pathlib import Path
from typing import Any

import openziti
from uvicorn import config, server
from uvicorn.lifespan.on import LifespanOn
from uvicorn.protocols.http.httptools_impl import HttpToolsProtocol as UvHttpToolsProtocol

from mrok.proxy.models import Identity
from mrok.types.proxy import ASGIApp

logger = logging.getLogger("mrok.proxy")

config.LIFESPAN["auto"] = "mrok.proxy.ziticorn:Lifespan"


class Lifespan(LifespanOn):
    def __init__(self, lf_config: config.Config) -> None:
        super().__init__(lf_config)
        self.logger = logging.getLogger("mrok.proxy")


class HttpToolsProtocol(UvHttpToolsProtocol):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger("mrok.proxy")
        self.access_logger = logging.getLogger("mrok.access")
        self.access_log = self.access_logger.hasHandlers()


class Server(server.Server):
    async def serve(self, sockets: list[socket.socket] | None = None) -> None:
        if not sockets:
            sockets = [self.config.bind_socket()]
        with self.capture_signals():
            await self._serve(sockets)


class BackendConfig(config.Config):
    def __init__(
        self,
        app: ASGIApp | Callable[..., Any] | str,
        identity_file: str | Path,
        ziti_load_timeout_ms: int = 5000,
        backlog: int = 2048,
        timeout_keep_alive: int = 5,
        limit_concurrency: int | None = None,
        limit_max_requests: int | None = None,
    ):
        self.identity_file = identity_file
        self.identity = Identity.load_from_file(self.identity_file)
        self.ziti_load_timeout_ms = ziti_load_timeout_ms
        super().__init__(
            app,
            loop="asyncio",
            http=HttpToolsProtocol,
            backlog=backlog,
            timeout_keep_alive=timeout_keep_alive,
            limit_concurrency=limit_concurrency,
            limit_max_requests=limit_max_requests,
        )

    def bind_socket(self) -> socket.socket:
        logger.info(
            "Connect to Ziti service "
            f"'{self.identity.mrok.extension} ({self.identity.mrok.instance})'"
        )

        ctx, err = openziti.load(str(self.identity_file), timeout=self.ziti_load_timeout_ms)
        if err != 0:
            raise RuntimeError(f"Failed to load Ziti identity from {self.identity_file}: {err}")

        sock = ctx.bind(self.identity.mrok.extension)
        sock.listen(self.backlog)
        logger.info(f"listening on ziti service {self.identity.mrok.extension} for connections")
        return sock

    def configure_logging(self) -> None:
        return



import abc
import logging

from httpcore import AsyncConnectionPool, Request

from mrok.proxy.exceptions import ProxyError
from mrok.proxy.stream import ASGIRequestBodyStream
from mrok.types.proxy import ASGIReceive, ASGISend, Scope

logger = logging.getLogger("mrok.proxy")


HOP_BY_HOP_HEADERS = [
    b"connection",
    b"keep-alive",
    b"proxy-authenticate",
    b"proxy-authorization",
    b"te",
    b"trailers",
    b"transfer-encoding",
    b"upgrade",
]


class ProxyAppBase(abc.ABC):
    def __init__(
        self,
        *,
        max_connections: int | None = 10,
        max_keepalive_connections: int | None = None,
        keepalive_expiry: float | None = None,
        retries: int = 0,
    ) -> None:
        self._pool = self.setup_connection_pool(
            max_connections=max_connections,
            max_keepalive_connections=max_keepalive_connections,
            keepalive_expiry=keepalive_expiry,
            retries=retries,
        )

    @abc.abstractmethod
    def setup_connection_pool(
        self,
        max_connections: int | None,
        max_keepalive_connections: int | None,
        keepalive_expiry: float | None,
        retries: int,
    ) -> AsyncConnectionPool:
        raise NotImplementedError()

    @abc.abstractmethod
    def get_upstream_base_url(self, scope: Scope) -> str:
        raise NotImplementedError()

    async def __call__(self, scope: Scope, receive: ASGIReceive, send: ASGISend) -> None:
        if scope.get("type") == "lifespan":
            return

        if scope.get("type") != "http":
            await self.send_error_response(scope, send, 500, "Unsupported")
            return

        try:
            base_url = self.get_upstream_base_url(scope)
            if base_url.endswith("/"):  # pragma: no cover
                base_url = base_url[:-1]
            full_path = self._format_path(scope)
            url = f"{base_url}{full_path}"
            method = scope.get("method", "GET").encode()
            headers = self._prepare_headers(scope)

            body_stream = ASGIRequestBodyStream(receive)

            request = Request(
                method=method,
                url=url,
                headers=headers,
                content=body_stream,
            )
            response = await self._pool.handle_async_request(request)
            logger.debug(f"connection pool status: {self._pool}")
            response_headers = []
            for k, v in response.headers:
                if k.lower() not in HOP_BY_HOP_HEADERS:
                    response_headers.append((k, v))

            await send(
                {
                    "type": "http.response.start",
                    "status": response.status,
                    "headers": response_headers,
                }
            )

            async for chunk in response.stream:  # type: ignore[union-attr]
                await send(
                    {
                        "type": "http.response.body",
                        "body": chunk,
                        "more_body": True,
                    }
                )

            await send({"type": "http.response.body", "body": b"", "more_body": False})
            await response.aclose()

        except ProxyError as pe:
            await self.send_error_response(scope, send, pe.http_status, pe.message)

        except Exception:
            logger.exception("Unexpected error in forwarder")
            await self.send_error_response(scope, send, 502, "Bad Gateway")

    async def send_error_response(
        self,
        scope: Scope,
        send: ASGISend,
        http_status: int,
        body: str,
        headers: list[tuple[bytes, bytes]] | None = None,
    ):
        headers = headers or [(b"content-type", b"text/plain")]
        try:
            await send({"type": "http.response.start", "status": http_status, "headers": headers})
            await send({"type": "http.response.body", "body": body.encode()})
        except Exception as e:  # pragma: no cover
            logger.error(f"Cannot send error response: {e}")

    def _prepare_headers(self, scope: Scope) -> list[tuple[bytes, bytes]]:
        headers: list[tuple[bytes, bytes]] = []
        scope_headers = scope.get("headers", [])

        for k, v in scope_headers:
            if k.lower() not in HOP_BY_HOP_HEADERS:
                headers.append((k, v))

        self._merge_x_forwarded(headers, scope)

        return headers

    def _find_header(self, headers: list[tuple[bytes, bytes]], name: bytes) -> int | None:
        """Return index of header `name` in `headers`, or None if missing."""
        lname = name.lower()
        for i, (k, _) in enumerate(headers):
            if k.lower() == lname:
                return i
        return None

    def _merge_x_forwarded(self, headers: list[tuple[bytes, bytes]], scope: Scope) -> None:
        client = scope.get("client")
        if client:
            client_ip = client[0].encode()
            idx = self._find_header(headers, b"x-forwarded-for")
            if idx is None:
                headers.append((b"x-forwarded-for", client_ip))
            else:
                k, v = headers[idx]
                headers[idx] = (k, v + b", " + client_ip)

        server = scope.get("server")
        if server:
            if self._find_header(headers, b"x-forwarded-host") is None:
                headers.append((b"x-forwarded-host", server[0].encode()))
            if server[1] and self._find_header(headers, b"x-forwarded-port") is None:
                headers.append((b"x-forwarded-port", str(server[1]).encode()))

        # Always set the protocol to https for upstream
        idx_proto = self._find_header(headers, b"x-forwarded-proto")
        if idx_proto is None:
            headers.append((b"x-forwarded-proto", b"https"))
        else:
            k, _ = headers[idx_proto]
            headers[idx_proto] = (k, b"https")

    def _format_path(self, scope: Scope) -> str:
        raw_path = scope.get("raw_path")
        if raw_path:
            return raw_path.decode()
        q = scope.get("query_string", b"")
        path = scope.get("path", "/")
        path_qs = path
        if q:
            path_qs += "?" + q.decode()
        return path_qs



from pytest_mock import MockerFixture

from mrok.frontend.main import run


def test_run(mocker: MockerFixture):
    mocker.patch(
        "mrok.frontend.main.get_logging_config",
        return_value={"logging": "config"},
    )

    m_healthcheck_mw_app = mocker.MagicMock()
    m_healthcheck_mw_app_ctor = mocker.patch(
        "mrok.frontend.main.HealthCheckMiddleware", return_value=m_healthcheck_mw_app
    )

    m_asgi_app = mocker.MagicMock()
    m_asgi_app_ctor = mocker.patch("mrok.frontend.main.FrontendProxyApp", return_value=m_asgi_app)
    m_app = mocker.MagicMock()
    m_standalone_app = mocker.patch("mrok.frontend.main.StandaloneApplication", return_value=m_app)
    run("my-identity.json", "localhost", 2423, 4, 1001, 323, 99.5)
    m_standalone_app.assert_called_once_with(
        m_healthcheck_mw_app,
        {
            "bind": "localhost:2423",
            "workers": 4,
            "worker_class": "mrok.frontend.main.MrokUvicornWorker",
            "logconfig_dict": {"logging": "config"},
            "reload": False,
        },
    )
    m_app.run.assert_called_once()
    m_asgi_app_ctor.assert_called_once_with(
        "my-identity.json",
        max_connections=1001,
        max_keepalive_connections=323,
        keepalive_expiry=99.5,
    )
    m_healthcheck_mw_app_ctor.assert_called_once_with(m_asgi_app)



import json

from mrok.frontend.utils import get_target_name
from mrok.types.proxy import ASGIApp, ASGIReceive, ASGISend, Scope


class HealthCheckMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: ASGIReceive, send: ASGISend):
        if scope["type"] == "http" and scope["path"] == "/healthcheck":
            target = get_target_name(
                {k.decode("latin1"): v.decode("latin1") for k, v in scope.get("headers", {})}
            )

            if not target:
                await send(
                    {
                        "type": "http.response.start",
                        "status": 200,
                        "headers": [
                            [b"content-type", b"application/json"],
                        ],
                    }
                )
                await send(
                    {
                        "type": "http.response.body",
                        "body": json.dumps({"status": "healthy"}).encode("utf-8"),
                    }
                )
                return

        await self.app(scope, receive, send)



from collections.abc import Awaitable, Callable, Iterable
from pathlib import PurePath
from typing import Any, Protocol

import zmq
from textual.app import App
from textual.pilot import Pilot

from mrok.conf import Settings
from mrok.types.proxy import ASGIReceive, ASGISend, Message

ZMQPublisher = tuple[zmq.Socket, int]


class SnapCompare(Protocol):
    def __call__(
        self,
        app: str | PurePath | App,
        press: Iterable[str] = (),
        terminal_size: tuple[int, int] = (80, 24),
        run_before: Callable[[Pilot], Awaitable[None] | None] | None = None,
    ) -> bool: ...


SettingsFactory = Callable[..., Settings]


class ReceiveFactory(Protocol):
    def __call__(self, messages: list[Message] | None = None) -> ASGIReceive: ...


SendFactory = Callable[[list[Message]], ASGISend]


class StatusEventFactory(Protocol):
    def __call__(
        self,
        process_cpu: float = 55.1,
        process_mem: float = 21.2,
        requests_rps: int = 123,
        requests_total: int = 1000,
        requests_successful: int = 10,
        requests_failed: int = 30,
        bytes_in: int = 1000,
        bytes_out: int = 2000,
    ) -> dict[str, Any]: ...


class ResponseEventFactory(Protocol):
    def __call__(
        self,
        method: str = "GET",
        url: str = "/test",
        request_headers: dict[str, str] | None = None,
        request_querystring: bytes | None = None,
        request_body: bytes | None = None,
        request_truncated: bool = False,
        response_headers: dict[str, str] | None = None,
        response_status: int = 200,
        response_body: bytes | None = b'{"test": "json"}',
        response_truncated: bool = False,
        duration: float = 10.3,
    ) -> dict[str, Any]: ...



from pathlib import Path
from typing import Annotated

import typer

from mrok.agent import sidecar
from mrok.cli.utils import number_of_workers

default_workers = number_of_workers()


def register(app: typer.Typer) -> None:
    @app.command("sidecar")
    def run_sidecar(
        identity_file: Path = typer.Argument(
            ...,
            help="Identity json file",
        ),
        target: Path = typer.Argument(
            ...,
            help="Target service (host:port or path to unix domain socket)",
        ),
        ziti_load_timeout_ms: Annotated[
            int,
            typer.Option(
                "--ziti-load-timeout-ms",
                help="Timeout (ms) waiting for Ziti to load.",
                show_default=True,
            ),
        ] = 5000,
        server_workers: Annotated[
            int,
            typer.Option(
                "--server-workers",
                "-w",
                help="Number of workers.",
                show_default=True,
            ),
        ] = default_workers,
        server_backlog: Annotated[
            int,
            typer.Option(
                "--server-backlog",
                help="TCP socket listen backlog.",
                show_default=True,
            ),
        ] = 2048,
        server_timeout_keep_alive: Annotated[
            int,
            typer.Option(
                "--server-timeout-keep-alive",
                help="Seconds to keep idle HTTP connections open.",
                show_default=True,
            ),
        ] = 5,
        server_limit_concurrency: Annotated[
            int | None,
            typer.Option(
                "--server-limit-concurrency",
                help="Maximum number of concurrent requests per worker.",
                show_default=True,
            ),
        ] = None,
        server_limit_max_requests: Annotated[
            int | None,
            None,
            typer.Option(
                "--server-limit-max-requests",
                help="Restart a worker after handling this many requests.",
                show_default=True,
            ),
        ] = None,
        events_publishers_port: Annotated[
            int,
            typer.Option(
                "--events-publishers-port",
                "-p",
                help="TCP port where the mrok agent should connect to publish events.",
                show_default=True,
            ),
        ] = 50000,
        events_subscribers_port: Annotated[
            int,
            typer.Option(
                "--events-subscribers-port",
                help=(
                    "TCP port where the mrok agent should listen for incoming subscribers "
                    "connections for listening to events."
                ),
                show_default=True,
            ),
        ] = 50001,
        events_metrics_collect_interval: Annotated[
            float,
            typer.Option(
                "--events-metrics-collect-interval",
                help="Interval in seconds between events metrics collect.",
                show_default=True,
            ),
        ] = 5.0,
        no_events: Annotated[
            bool,
            typer.Option(
                "--no-events",
                help="Disable events. Default: False",
                show_default=True,
            ),
        ] = False,
        upstream_max_connections: Annotated[
            int,
            typer.Option(
                "--upstream-max-connections",
                help=(
                    "The maximum number of concurrent HTTP connections to the target service that "
                    "the pool should allow. Any attempt to send a request on "
                    "a pool that would exceed this amount will block until a connection "
                    "is available."
                ),
                show_default=True,
            ),
        ] = 10,
        upstream_max_keepalive_connections: Annotated[
            int | None,
            typer.Option(
                "--upstream-max-keepalive-connections",
                help=(
                    "The maximum number of idle HTTP connections "
                    "that will be maintained in the connection pool to the "
                    "target service."
                ),
                show_default=True,
            ),
        ] = None,
        upstream_keepalive_expiry: Annotated[
            float | None,
            typer.Option(
                "--upstream_keepalive_expiry",
                help=(
                    "The duration in seconds that an idle HTTP connection to the target service"
                    "may be maintained for before being expired from the pool."
                ),
                show_default=True,
            ),
        ] = None,
        upstream_max_connect_retries: Annotated[
            int,
            typer.Option(
                "--upstream-max-connect-retries",
                help=(
                    "The duration in seconds that an idle HTTP connection to the target service"
                    "may be maintained for before being expired from the pool."
                ),
                show_default=True,
            ),
        ] = 0,
    ):
        """Run a Sidecar Proxy to expose a web application through OpenZiti."""
        if ":" in str(target):
            host, port = str(target).split(":", 1)
            target_addr = (host or "127.0.0.1", int(port))
        else:
            target_addr = str(target)  # type: ignore

        sidecar.run(
            str(identity_file),
            target_addr,
            ziti_load_timeout_ms=ziti_load_timeout_ms,
            server_workers=server_workers,
            server_backlog=server_backlog,
            server_timeout_keep_alive=server_timeout_keep_alive,
            server_limit_concurrency=server_limit_concurrency,
            server_limit_max_requests=server_limit_max_requests,
            events_enabled=not no_events,
            events_publishers_port=events_publishers_port,
            events_subscribers_port=events_subscribers_port,
            events_metrics_collect_interval=events_metrics_collect_interval,
            upstream_max_connections=upstream_max_connections,
            upstream_max_keepalive_connections=upstream_max_keepalive_connections,
            upstream_keepalive_expiry=upstream_keepalive_expiry,
            upstream_max_connect_retries=upstream_max_connect_retries,
        )



import asyncio
from collections.abc import Callable
from http import HTTPStatus
from typing import Any

import pytest
from httpcore import Request
from pytest_mock import MockerFixture

from mrok.proxy.app import HOP_BY_HOP_HEADERS, ProxyAppBase
from mrok.proxy.exceptions import ProxyError
from mrok.types.proxy import ASGIReceive, ASGISend, Message
from tests.types import ReceiveFactory, SendFactory


class _DummyResponse:
    def __init__(self, *, headers=None, status: int = 200, chunks: list[bytes] | None = None):
        self.headers = headers or [(b"content-type", b"text/plain")]
        self.status = status
        self._chunks = chunks or [b"ok"]

    @property
    def stream(self):
        async def _gen():
            await asyncio.sleep(0)
            for c in self._chunks:
                yield c

        return _gen()

    async def aclose(self) -> None:
        return None


def _find_header(hdrs, name: bytes):
    for k, v in hdrs:
        if k.lower() == name:
            return v
    return None


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "incoming_xf",
    [
        [],
        [
            (b"x-forwarded-for", b"1.2.3.4"),
            (b"x-forwarded-host", b"orig-host"),
            (b"x-forwarded-port", b"123"),
            (b"x-forwarded-proto", b"http"),
        ],
    ],
)
@pytest.mark.parametrize(
    "hop_by_hop_headers",
    [
        [],
        HOP_BY_HOP_HEADERS,
    ],
)
async def test_app_success(
    incoming_xf: list[tuple[bytes, bytes]],
    hop_by_hop_headers: list[bytes],
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
) -> None:
    sent: list[Message] = []

    captured: dict[str, Any] = {}
    resp_headers = [(b"content-type", b"text/plain")]
    resp_headers.extend([(hhh_name, b"value") for hhh_name in hop_by_hop_headers])

    class Pool:
        async def handle_async_request(self, req):
            captured["req"] = req
            return _DummyResponse(
                headers=resp_headers,
                status=200,
                chunks=[b"one", b"two"],
            )

    pool = Pool()

    class ProxyApp(ProxyAppBase):
        def setup_connection_pool(
            self,
            max_connections=1000,
            max_keepalive_connections=10,
            keepalive_expiry=120,
            retries=0,
        ):
            return pool

        def get_upstream_base_url(self, scope):
            return "http://upstream"

    app = ProxyApp()

    scope_headers: list[tuple[bytes, bytes]] = []
    scope_headers.extend(incoming_xf)
    scope_headers.extend([(hhh_name, b"value") for hhh_name in hop_by_hop_headers])

    scope = {
        "type": "http",
        "method": "POST",
        "raw_path": b"/foo",
        "query_string": b"param=value",
        "headers": scope_headers,
        "scheme": "https",
        "client": ("127.0.0.1", 12345),
        "server": ("localhost", 8000),
    }

    receive = receive_factory()
    send = send_factory(sent)

    await app(scope, receive, send)

    assert sent[0]["type"] == "http.response.start"
    assert sent[0]["status"] == 200

    returned_headers = sent[0]["headers"]
    for k, _ in returned_headers:
        assert k.lower() not in HOP_BY_HOP_HEADERS

    # body chunks
    assert sent[1]["body"] == b"one"
    assert sent[1]["more_body"] is True
    assert sent[2]["body"] == b"two"
    assert sent[2]["more_body"] is True
    # final empty message
    assert sent[3]["body"] == b""
    assert sent[3]["more_body"] is False

    req: Request = captured["req"]

    assert bytes(req.url) == b"http://upstream/foo"
    for header in req.headers:
        assert header[0] not in HOP_BY_HOP_HEADERS

    client_ip = b"127.0.0.1"
    xf_for = _find_header(req.headers, b"x-forwarded-for")
    if incoming_xf:
        assert xf_for == b"1.2.3.4, " + client_ip
        assert _find_header(req.headers, b"x-forwarded-host") == b"orig-host"
        assert _find_header(req.headers, b"x-forwarded-port") == b"123"
    else:
        assert xf_for == client_ip
        assert _find_header(req.headers, b"x-forwarded-host") == b"localhost"
        assert _find_header(req.headers, b"x-forwarded-port") == b"8000"

    assert _find_header(req.headers, b"x-forwarded-proto") == b"https"


@pytest.mark.asyncio
async def test_proxyerror_returns_custom_status_and_message(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
) -> None:
    sent: list[Message] = []

    class App(ProxyAppBase):
        def setup_connection_pool(self, *a, **k):
            return None

        def get_upstream_base_url(self, scope):
            raise ProxyError(HTTPStatus.IM_A_TEAPOT, "short and stout")

    app = App()
    scope = {"type": "http", "path": "/"}
    receive = receive_factory()
    send = send_factory(sent)

    await app(scope, receive, send)

    assert sent[0]["type"] == "http.response.start"
    assert sent[0]["status"] == HTTPStatus.IM_A_TEAPOT
    assert sent[1]["body"] == b"short and stout"


@pytest.mark.asyncio
async def test_generic_exception_produces_502(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
) -> None:
    sent: list[Message] = []

    class Pool:
        async def handle_async_request(self, req):
            raise RuntimeError("boom")

    class ProxyApp(ProxyAppBase):
        def setup_connection_pool(self, *a, **k):
            return Pool()

        def get_upstream_base_url(self, scope):
            return "http://upstream"

    app = ProxyApp()
    scope = {"type": "http", "path": "/", "method": "GET"}
    receive = receive_factory()
    send = send_factory(sent)

    await app(scope, receive, send)

    assert sent[0]["type"] == "http.response.start"
    assert sent[0]["status"] == 502
    assert sent[1]["body"] == b"Bad Gateway"


@pytest.mark.asyncio
async def test_lifespan_scope_no_send(
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
) -> None:
    class ProxyApp(ProxyAppBase):
        def setup_connection_pool(
            self,
            max_connections=1000,
            max_keepalive_connections=10,
            keepalive_expiry=120,
            retries=0,
        ):
            pass

        def get_upstream_base_url(self, scope):
            pass

    sends: list[Message] = []
    app = ProxyApp()
    scope = {"type": "lifespan"}
    receive = receive_factory()
    send = send_factory(sends)

    await app(scope, receive, send)
    assert sends == []


@pytest.mark.asyncio
async def test_non_http_scope_sends_unsupported(
    receive_factory: Callable[[], ASGIReceive],
    send_factory: Callable[[list[dict]], ASGISend],
) -> None:
    class ProxyApp(ProxyAppBase):
        def setup_connection_pool(
            self,
            max_connections=1000,
            max_keepalive_connections=10,
            keepalive_expiry=120,
            retries=0,
        ):
            pass

        def get_upstream_base_url(self, scope):
            pass

    sends: list[dict] = []
    app = ProxyApp()
    scope = {"type": "websocket"}
    receive = receive_factory()
    send = send_factory(sends)

    await app(scope, receive, send)

    assert len(sends) == 2
    assert sends[0]["type"] == "http.response.start"
    assert sends[0]["status"] == 500
    assert sends[1]["type"] == "http.response.body"
    assert sends[1]["body"] == b"Unsupported"



from http import HTTPStatus
from pathlib import Path
from typing import Any

from httpcore import AsyncConnectionPool
from jinja2 import Environment, FileSystemLoader, select_autoescape

from mrok.conf import get_settings
from mrok.frontend.utils import get_target_name, parse_accept_header
from mrok.proxy.app import ProxyAppBase
from mrok.proxy.backend import AIOZitiNetworkBackend
from mrok.proxy.exceptions import InvalidTargetError
from mrok.types.proxy import ASGISend, Scope

ERROR_TEMPLATE_FORMATS = {
    "application/json": "json",
    "text/html": "html",
}


class FrontendProxyApp(ProxyAppBase):
    def __init__(
        self,
        identity_file: str,
        *,
        max_connections: int | None = 10,
        max_keepalive_connections: int | None = None,
        keepalive_expiry: float | None = None,
        retries: int = 0,
    ):
        self._identity_file = identity_file
        self._jinja_env_cache: dict[Path, Environment] = {}
        self._templates_by_error = get_settings().frontend.get("errors", {})
        super().__init__(
            max_connections=max_connections,
            max_keepalive_connections=max_keepalive_connections,
            keepalive_expiry=keepalive_expiry,
            retries=retries,
        )

    def setup_connection_pool(
        self,
        max_connections: int | None,
        max_keepalive_connections: int | None,
        keepalive_expiry: float | None,
        retries: int,
    ) -> AsyncConnectionPool:
        return AsyncConnectionPool(
            max_connections=max_connections,
            max_keepalive_connections=max_keepalive_connections,
            keepalive_expiry=keepalive_expiry,
            retries=retries,
            network_backend=AIOZitiNetworkBackend(self._identity_file),
        )

    def get_upstream_base_url(self, scope: Scope) -> str:
        target = get_target_name(
            {k.decode("latin1"): v.decode("latin1") for k, v in scope.get("headers", {})}
        )
        if not target:
            raise InvalidTargetError()

        return f"http://{target.lower()}"

    async def send_error_response(
        self,
        scope: Scope,
        send: ASGISend,
        http_status: int,
        body: str,
        headers: list[tuple[bytes, bytes]] | None = None,
    ):
        request_headers = {
            k.decode("latin1"): v.decode("latin1") for k, v in scope.get("headers", {})
        }
        accept_header = request_headers.get("accept")
        if not (accept_header and str(http_status) in self._templates_by_error):
            return await super().send_error_response(scope, send, http_status, body)

        available_templates = self._templates_by_error[str(http_status)]

        media_types = parse_accept_header(accept_header)
        for media_type in media_types:
            template_format = ERROR_TEMPLATE_FORMATS.get(media_type)
            if template_format and template_format in available_templates:
                template_path = available_templates[template_format]
                rendered = await self._render_error_template(
                    Path(template_path), scope, http_status, body
                )
                return await super().send_error_response(
                    scope,
                    send,
                    http_status,
                    rendered,
                    headers=[(b"content-type", media_type.encode("latin-1"))],
                )

        return await super().send_error_response(scope, send, http_status, body)

    async def _render_error_template(
        self, template_path: Path, scope: Scope, http_status: int, body: str
    ) -> str:
        env = self._get_jinja_env(template_path)
        template = env.get_template(template_path.name)
        status_title = HTTPStatus(http_status).name.replace("_", " ").title()
        context = {
            "status": http_status,
            "status_title": status_title,
            "body": body,
            "request": self._extract_request_context(scope),
        }

        return await template.render_async(context)

    def _extract_request_context(self, scope: Scope) -> dict[str, Any]:
        headers = {k.decode("latin-1"): v.decode("latin-1") for k, v in scope.get("headers", [])}

        return {
            "method": scope.get("method"),
            "path": scope.get("path"),
            "raw_path": scope.get("raw_path", b"").decode("latin-1"),
            "query_string": scope.get("query_string", b"").decode("latin-1"),
            "scheme": scope.get("scheme"),
            "headers": headers,
            "client": scope.get("client"),
            "server": scope.get("server"),
            "http_version": scope.get("http_version"),
        }

    def _get_jinja_env(self, template_path: Path) -> Environment:
        template_dir = template_path.parent

        if template_dir not in self._jinja_env_cache:
            self._jinja_env_cache[template_dir] = Environment(
                loader=FileSystemLoader(str(template_dir)),
                autoescape=select_autoescape(
                    enabled_extensions=("html", "xml"),
                    default_for_string=False,
                ),
                enable_async=True,
            )

        return self._jinja_env_cache[template_dir]



import pytest
from jinja2 import Template
from pytest_mock import MockerFixture

from mrok.frontend.app import FrontendProxyApp
from mrok.proxy.app import ProxyAppBase
from mrok.proxy.exceptions import InvalidTargetError
from tests.types import SettingsFactory


def test_init(
    mocker: MockerFixture,
):
    m_async_pool_ctor = mocker.patch("mrok.frontend.app.AsyncConnectionPool")
    m_ziti_backend = mocker.MagicMock()
    m_ziti_backend_ctor = mocker.patch(
        "mrok.frontend.app.AIOZitiNetworkBackend", return_value=m_ziti_backend
    )
    FrontendProxyApp(
        "my-identity-file",
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
    )

    m_async_pool_ctor.assert_called_once_with(
        max_connections=5000,
        max_keepalive_connections=5,
        keepalive_expiry=60,
        retries=1,
        network_backend=m_ziti_backend,
    )
    m_ziti_backend_ctor.assert_called_once_with("my-identity-file")


@pytest.mark.parametrize(
    ("header", "expected"),
    [
        ((b"x-forwarded-host", b"ext-9284-9038.extdomain"), "ext-9284-9038"),
        ((b"host", b"ext-9284-9038.extdomain"), "ext-9284-9038"),
        ((b"x-forwarded-host", b"eXt-9284-9038.extdomain"), "ext-9284-9038"),
        ((b"host", b"EXT-9284-9038.extdomain"), "ext-9284-9038"),
        ((b"x-forwarded-host", b"ins-9284-9038-0001.extdomain"), "ins-9284-9038-0001"),
        ((b"host", b"ins-9284-9038-0001.extdomain"), "ins-9284-9038-0001"),
        ((b"x-forwarded-host", b"inS-9284-9038-0001.extdomain"), "ins-9284-9038-0001"),
        ((b"host", b"INS-9284-9038-0001.extdomain"), "ins-9284-9038-0001"),
        ((b"x-forwarded-host", b"ext-9284-9038.extdomain:1234"), "ext-9284-9038"),
        ((b"host", b"ext-9284-9038.extdomain:1234"), "ext-9284-9038"),
        ((b"x-forwarded-host", b"eXt-9284-9038.extdomain:1234"), "ext-9284-9038"),
        ((b"host", b"EXT-9284-9038.extdomain:1234"), "ext-9284-9038"),
        ((b"x-forwarded-host", b"ins-9284-9038-0001.extdomain:1234"), "ins-9284-9038-0001"),
        ((b"host", b"ins-9284-9038-0001.extdomain:1234"), "ins-9284-9038-0001"),
        ((b"x-forwarded-host", b"inS-9284-9038-0001.extdomain:1234"), "ins-9284-9038-0001"),
        ((b"host", b"INS-9284-9038-0001.extdomain:1234"), "ins-9284-9038-0001"),
    ],
)
def test_get_upstream_base_url(
    mocker: MockerFixture,
    header: tuple[bytes, bytes],
    expected: str,
):
    mocker.patch("mrok.frontend.utils.get_frontend_domain", return_value=".extdomain")

    app = FrontendProxyApp("my-identity")
    assert app.get_upstream_base_url({"headers": [header]}) == f"http://{expected}"


@pytest.mark.parametrize(
    "header",
    [
        (b"x-forwarded-host", b"ext-9284.extdomain"),
        (b"host", b"ext-9284.extdomain"),
        (b"x-forwarded-host", b"ins-9292-9292-1.extdomain"),
        (b"host", b"ins-9292-9292-1.extdomain"),
        (b"x-forwarded-host", b"whatever.extdomain"),
        (b"host", b"whatever.extdomain"),
        (b"x-forwarded-host", b"ext-9284-9038.otherdomain"),
        (b"host", b"ext-9284-9038.otherdomain"),
        (b"x-forwarded-host", b"ins-9284-9038-0001.otherdomain"),
        (b"host", b"ins-9284-9038-0001.otherdomain"),
    ],
)
def test_get_upstream_base_url_invalid_target(
    mocker: MockerFixture,
    header: tuple[bytes, bytes],
):
    mocker.patch("mrok.frontend.utils.get_frontend_domain", return_value=".extdomain")

    app = FrontendProxyApp("my-identity")
    with pytest.raises(InvalidTargetError):
        app.get_upstream_base_url({"headers": [header]})


@pytest.mark.parametrize(
    "accept_header",
    [
        "text/html, application/json",
        "text/html, application/json, text/plain",
        "text/html;q=1.0, application/json;q=0.9, */*;q=0.8",
        "text/html;q=0.9, application/json;q=0.5",
        "text/html, application/*;q=0.9, application/json;q=0.8",
        "text/html, application/json;q=1.0",
        "text/html,application/xhtml+xml,application/json;q=0.9,image/webp,*/*;q=0.8",
        "application/json;q=0.1, text/html;q=1.0",
    ],
)
@pytest.mark.asyncio
async def test_html_error(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
    ziti_frontend_error_template_html_file: str,
    ziti_frontend_error_template_html: str,
    accept_header: str,
):
    template = Template(ziti_frontend_error_template_html)
    settings = settings_factory(
        frontend={
            "domain": "ext.mrok.test",
            "errors": {
                "502": {
                    "html": ziti_frontend_error_template_html_file,
                }
            },
        }
    )
    mocker.patch("mrok.frontend.app.get_settings", return_value=settings)
    m_send_error = mocker.patch.object(ProxyAppBase, "send_error_response")
    m_send = mocker.AsyncMock()
    scope = {
        "headers": [
            (b"accept", accept_header.encode("latin-1")),
        ],
    }

    app = FrontendProxyApp("my-identity")

    await app.send_error_response(
        {
            "headers": [
                (b"accept", accept_header.encode("latin-1")),
            ],
        },
        m_send,
        502,
        "bad gateway",
    )
    m_send_error.assert_awaited_once_with(
        scope,
        m_send,
        502,
        template.render({"status": 502, "body": "bad gateway"}),
        headers=[
            (b"content-type", b"text/html"),
        ],
    )


@pytest.mark.asyncio
async def test_html_error_no_html_template(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
):
    settings = settings_factory(
        frontend={
            "domain": "ext.mrok.test",
            "errors": {
                "502": {
                    "xml": "/path/to/xml/template",
                }
            },
        }
    )
    mocker.patch("mrok.frontend.app.get_settings", return_value=settings)
    m_send_error = mocker.patch.object(ProxyAppBase, "send_error_response")
    m_send = mocker.AsyncMock()
    scope = {
        "headers": [
            (b"accept", b"text/html"),
        ],
    }

    app = FrontendProxyApp("my-identity")

    await app.send_error_response(
        {
            "headers": [
                (b"accept", b"text/html"),
            ],
        },
        m_send,
        502,
        "bad gateway",
    )
    m_send_error.assert_awaited_once_with(
        scope,
        m_send,
        502,
        "bad gateway",
    )


@pytest.mark.asyncio
async def test_html_error_no_accept(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
    ziti_frontend_error_template_html_file: str,
):
    settings = settings_factory(
        frontend={
            "domain": "ext.mrok.test",
            "errors": {
                "502": {
                    "html": ziti_frontend_error_template_html_file,
                }
            },
        }
    )
    mocker.patch("mrok.frontend.app.get_settings", return_value=settings)
    m_send_error = mocker.patch.object(ProxyAppBase, "send_error_response")
    m_send = mocker.AsyncMock()
    scope = {
        "headers": [
            (b"host", b"localhost"),
        ],
    }

    app = FrontendProxyApp("my-identity")

    await app.send_error_response(
        {
            "headers": [
                (b"host", b"localhost"),
            ],
        },
        m_send,
        502,
        "bad gateway",
    )
    m_send_error.assert_awaited_once_with(
        scope,
        m_send,
        502,
        "bad gateway",
    )


@pytest.mark.parametrize(
    "accept_header",
    [
        "application/json",
        "application/json, text/javascript, */*",
        "text/html;q=0.9, application/json;q=1.0, text/plain;q=0.5",
        "application/json;q=0.8, application/xml;q=0.8",
        "application/json, application/*;q=0.9, */*;q=0.8",
        "application/json, text/plain, */*;q=0.01",
        "application/json;q=0.1, text/html;q=0",
        "text/html,application/xhtml+xml,application/xml;q=0.9,application/json;q=1.0",
    ],
)
@pytest.mark.asyncio
async def test_json_error(
    mocker: MockerFixture,
    settings_factory: SettingsFactory,
    ziti_frontend_error_template_json_file: str,
    ziti_frontend_error_template_json: str,
    accept_header: str,
):
    template = Template(ziti_frontend_error_template_json)
    settings = settings_factory(
        frontend={
            "domain": "ext.mrok.test",
            "errors": {
                "502": {
                    "json": ziti_frontend_error_template_json_file,
                }
            },
        }
    )
    mocker.patch("mrok.frontend.app.get_settings", return_value=settings)
    m_send_error = mocker.patch.object(ProxyAppBase, "send_error_response")
    m_send = mocker.AsyncMock()
    scope = {
        "headers": [
            (b"accept", accept_header.encode("latin-1")),
        ],
    }

    app = FrontendProxyApp("my-identity")

    await app.send_error_response(
        {
            "headers": [
                (b"accept", accept_header.encode("latin-1")),
            ],
        },
        m_send,
        502,
        "bad gateway",
    )
    m_send_error.assert_awaited_once_with(
        scope,
        m_send,
        502,
        template.render({"status": 502, "body": "bad gateway"}),
        headers=[
            (b"content-type", b"application/json"),
        ],
    )



import pytest
from pytest_mock import MockerFixture

from mrok.proxy.metrics import MetricsCollector, get_process_metrics
from mrok.proxy.models import ProcessMetrics


@pytest.mark.asyncio
async def test_get_process_metrics(
    mocker: MockerFixture,
):
    mocker.patch("mrok.proxy.metrics.os.getpid", return_value=1122)
    m_proc = mocker.MagicMock()
    m_proc.cpu_percent.side_effect = [Exception("oh"), 34.2]
    m_proc.memory_percent.return_value = 75.0
    m_proc_ctor = mocker.patch("mrok.proxy.metrics.psutil.Process", return_value=m_proc)

    metrics = await get_process_metrics(0.0001)
    assert metrics.cpu == 34.2
    assert metrics.mem == 75.0
    m_proc_ctor.assert_called_once_with(1122)


@pytest.mark.asyncio
async def test_worker_metrics_collector(
    mocker: MockerFixture,
):
    mocker.patch("mrok.proxy.metrics.time.perf_counter", side_effect=[0, 33, 75, 99])
    mocker.patch(
        "mrok.proxy.metrics.get_process_metrics", return_value=ProcessMetrics(cpu=7.3, mem=44.1)
    )
    collector = MetricsCollector("my-worker-id")
    begin = await collector.on_request_start({})
    await collector.on_request_body(23)
    await collector.on_request_body(32)
    await collector.on_response_start(200)
    await collector.on_response_chunk(11)
    await collector.on_response_chunk(13)
    await collector.on_request_end(begin, 200)

    begin = await collector.on_request_start({})
    await collector.on_request_body(11)
    await collector.on_request_body(4)
    await collector.on_response_start(500)
    await collector.on_request_end(begin, 500)

    snapshot = await collector.snapshot()

    assert snapshot.worker_id == "my-worker-id"
    assert snapshot.data_transfer.bytes_in == 70
    assert snapshot.data_transfer.bytes_out == 24
    assert snapshot.process.cpu == 7.3
    assert snapshot.process.mem == 44.1
    assert snapshot.requests.total == 2
    assert snapshot.requests.successful == 1
    assert snapshot.requests.failed == 1
    assert snapshot.requests.rps > 0
    assert snapshot.response_time.avg > 0
    assert snapshot.response_time.max > 0
    assert snapshot.response_time.min > 0
    assert snapshot.response_time.p50 > 0
    assert snapshot.response_time.p90 > 0
    assert snapshot.response_time.p99 > 0



from dynaconf import Dynaconf, LazySettings

type Settings = LazySettings

DEFAULT_SETTINGS = {
    "LOGGING": {
        "debug": False,
        "rich": False,
    },
    "FRONTEND": {
        "identity": "public",
        "mode": "zrok",
    },
    "ZITI": {
        "ssl_verify": False,
    },
    "CONTROLLER": {
        "pagination": {"limit": 50},
    },
    "IDENTIFIERS": {
        "extension": {
            "regex": "(?i)EXT-\\d{4}-\\d{4}",
            "format": "EXT-xxxx-yyyy",
            "example": "EXT-2000-1000",
        },
        "instance": {
            "regex": "(?i)INS-\\d{4}-\\d{4}-\\d{4}",
            "format": "INS-xxxx-yyyy-zzzz",
            "example": "INS-2004-2000-3000",
        },
    },
}

_settings = None


def get_settings() -> Settings:
    global _settings
    if not _settings:
        _settings = Dynaconf(
            envvar_prefix="MROK",
            settings_files=["settings.yaml", ".secrets.yaml"],
            merge_enabled=True,
        )
        _settings.configure(**DEFAULT_SETTINGS)
    return _settings



import shlex

from pytest_mock import MockerFixture
from typer.testing import CliRunner

from mrok.cli import app


def test_dev_console(
    mocker: MockerFixture,
):
    mocked_app = mocker.MagicMock()
    mocked_app_ctor = mocker.patch(
        "mrok.cli.commands.agent.dev.console.InspectorApp", return_value=mocked_app
    )
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("agent dev console -s 41234"),
    )
    assert result.exit_code == 0
    mocked_app_ctor.assert_called_once_with(41234)
    mocked_app.run.assert_called_once()


def test_dev_web(
    mocker: MockerFixture,
):
    mocked_server = mocker.MagicMock()
    mocked_server_ctor = mocker.patch(
        "mrok.cli.commands.agent.dev.web.InspectorServer", return_value=mocked_server
    )
    runner = CliRunner()

    # Run the command
    result = runner.invoke(
        app,
        shlex.split("agent dev web -p 9090 -s 41234"),
    )
    assert result.exit_code == 0
    mocked_server_ctor.assert_called_once_with(port=9090, subscriber_port=41234)
    mocked_server.serve.assert_called_once()



import json

import pytest
from pytest_mock import MockerFixture

from mrok.frontend.middleware import HealthCheckMiddleware


@pytest.mark.asyncio
async def test_healthcheck(
    mocker: MockerFixture,
):
    m_app = mocker.AsyncMock()
    m_receive = mocker.AsyncMock()
    m_send = mocker.AsyncMock()
    middleware = HealthCheckMiddleware(m_app)

    await middleware(
        {"type": "http", "path": "/healthcheck", "headers": [(b"host", b"127.0.0.1")]},
        m_receive,
        m_send,
    )

    assert m_send.mock_calls[0].args[0] == {
        "type": "http.response.start",
        "status": 200,
        "headers": [
            [b"content-type", b"application/json"],
        ],
    }

    assert m_send.mock_calls[1].args[0] == {
        "type": "http.response.body",
        "body": json.dumps({"status": "healthy"}).encode("utf-8"),
    }


@pytest.mark.asyncio
async def test_healthcheck_passtrhough(
    mocker: MockerFixture,
):
    m_app = mocker.AsyncMock()
    m_receive = mocker.AsyncMock()
    m_send = mocker.AsyncMock()
    mocker.patch("mrok.frontend.utils.get_frontend_domain", return_value=".extdomain")
    middleware = HealthCheckMiddleware(m_app)

    scope = {
        "type": "http",
        "path": "/healthcheck",
        "headers": [(b"host", b"ext-1234-5678.extdomain")],
    }

    await middleware(
        scope,
        m_receive,
        m_send,
    )

    m_app.assert_awaited_once_with(scope, m_receive, m_send)



from mrok.proxy.master import MasterBase
from mrok.types.proxy import ASGIApp


class ZiticornAgent(MasterBase):
    def __init__(
        self,
        app: ASGIApp | str,
        identity_file: str,
        ziti_load_timeout_ms: int = 5000,
        server_workers: int = 4,
        server_reload: bool = False,
        server_backlog: int = 2048,
        server_timeout_keep_alive: int = 5,
        server_limit_concurrency: int | None = None,
        server_limit_max_requests: int | None = None,
        events_publishers_port: int = 50000,
        events_subscribers_port: int = 5000,
        events_metrics_collect_interval: float = 5.0,
    ):
        super().__init__(
            identity_file,
            ziti_load_timeout_ms=ziti_load_timeout_ms,
            server_workers=server_workers,
            server_reload=server_reload,
            server_backlog=server_backlog,
            server_timeout_keep_alive=server_timeout_keep_alive,
            server_limit_concurrency=server_limit_concurrency,
            server_limit_max_requests=server_limit_max_requests,
            events_pub_port=events_publishers_port,
            events_sub_port=events_subscribers_port,
            events_metrics_collect_interval=events_metrics_collect_interval,
        )
        self.app = app

    def get_asgi_app(self):
        return self.app


def run(
    app: ASGIApp | str,
    identity_file: str,
    ziti_load_timeout_ms: int = 5000,
    server_workers: int = 4,
    server_reload: bool = False,
    server_backlog: int = 2048,
    server_timeout_keep_alive: int = 5,
    server_limit_concurrency: int | None = None,
    server_limit_max_requests: int | None = None,
    events_publishers_port: int = 50000,
    events_subscribers_port: int = 50001,
    events_metrics_collect_interval: float = 5.0,
):
    master = ZiticornAgent(
        app,
        identity_file,
        ziti_load_timeout_ms=ziti_load_timeout_ms,
        server_workers=server_workers,
        server_reload=server_reload,
        server_backlog=server_backlog,
        server_timeout_keep_alive=server_timeout_keep_alive,
        server_limit_concurrency=server_limit_concurrency,
        server_limit_max_requests=server_limit_max_requests,
        events_publishers_port=events_publishers_port,
        events_subscribers_port=events_subscribers_port,
        events_metrics_collect_interval=events_metrics_collect_interval,
    )
    master.run()



import asyncio

import pytest
from pytest_mock import MockerFixture

from mrok.proxy.middleware import CaptureMiddleware, MetricsMiddleware
from mrok.proxy.models import HTTPResponse
from mrok.types.proxy import Message
from tests.types import ReceiveFactory, SendFactory


@pytest.mark.asyncio
async def test_metrics(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
):
    class MockApp:
        async def __call__(self, scope, receive, send):
            await receive()
            await receive()
            await send({"type": "http.response.start", "status": 200})
            await send({"type": "http.response.body", "body": b"OK", "more_body": True})
            await send({"type": "http.response.body", "body": b"Mrok!", "more_body": False})

    m_app = MockApp()

    m_metrics = mocker.AsyncMock()
    m_metrics.on_request_start.return_value = 100

    sent: list[Message] = []
    receive = receive_factory(
        [
            {"type": "http.request", "body": b"Who are", "more_body": True},
            {"type": "http.request", "body": b"You!", "more_body": False},
        ]
    )
    send = send_factory(sent)

    middleware = MetricsMiddleware(m_app, m_metrics)
    await middleware({"type": "http"}, receive, send)

    m_metrics.on_request_start.assert_awaited_once_with({"type": "http"})
    assert m_metrics.on_request_body.mock_calls[0].args[0] == len(b"Who are")
    assert m_metrics.on_request_body.mock_calls[1].args[0] == len(b"You!")
    m_metrics.on_response_start.assert_awaited_once_with(200)
    assert m_metrics.on_response_chunk.mock_calls[0].args[0] == len(b"OK")
    assert m_metrics.on_response_chunk.mock_calls[1].args[0] == len(b"Mrok!")
    m_metrics.on_request_end.assert_awaited_once_with(100, 200)


@pytest.mark.asyncio
async def test_metrics_lifespan(
    mocker: MockerFixture,
):
    m_app = mocker.AsyncMock()
    m_metrics = mocker.AsyncMock()
    m_receive = mocker.AsyncMock()
    m_send = mocker.AsyncMock()

    middleware = MetricsMiddleware(m_app, m_metrics)
    await middleware({"type": "lifespan"}, m_receive, m_send)
    m_app.assert_awaited_once_with({"type": "lifespan"}, m_receive, m_send)


@pytest.mark.asyncio
async def test_capture(
    mocker: MockerFixture,
    receive_factory: ReceiveFactory,
    send_factory: SendFactory,
):
    mocker.patch("mrok.proxy.middleware.time.time", side_effect=[7, 25])

    class MockApp:
        async def __call__(self, scope, receive, send):
            await receive()
            await send(
                {
                    "type": "http.response.start",
                    "status": 200,
                    "headers": [(b"content-type", b"text/plain")],
                }
            )
            await send({"type": "http.response.body", "body": b"OK", "more_body": True})
            await send({"type": "http.response.body", "body": b"Mrok!", "more_body": False})

    m_app = MockApp()

    m_metrics = mocker.AsyncMock()
    m_metrics.on_request_start.return_value = 100

    sent: list[Message] = []
    receive = receive_factory(
        [
            {"type": "http.request", "body": b'{"my": "json"}', "more_body": False},
        ]
    )
    send = send_factory(sent)

    scope = {
        "type": "http",
        "method": "POST",
        "raw_path": b"/foo",
        "path": "/foo",
        "query_string": b"param=value",
        "headers": [(b"content-type", b"application/json")],
        "scheme": "https",
        "client": ("127.0.0.1", 12345),
        "server": ("localhost", 8000),
    }

    received_response: HTTPResponse | None = None

    async def on_response_complete(response: HTTPResponse):
        nonlocal received_response
        await asyncio.sleep(0)
        received_response = response

    middleware = CaptureMiddleware(m_app, on_response_complete)
    await middleware(scope, receive, send)
    await asyncio.sleep(0.1)

    assert received_response is not None
    assert received_response.request.method == "POST"
    assert received_response.request.url == "/foo"
    assert received_response.request.headers["content-type"] == "application/json"
    assert received_response.request.query_string == b"param=value"
    assert received_response.request.body == b'{"my": "json"}'
    assert received_response.status == 200
    assert received_response.headers["content-type"] == "text/plain"
    assert received_response.body == b"OKMrok!"
    assert received_response.duration == 18


@pytest.mark.asyncio
async def test_capture_lifespan(
    mocker: MockerFixture,
):
    m_app = mocker.AsyncMock()
    m_response_callback = mocker.AsyncMock()
    m_receive = mocker.AsyncMock()
    m_send = mocker.AsyncMock()

    middleware = CaptureMiddleware(m_app, m_response_callback)
    await middleware({"type": "lifespan"}, m_receive, m_send)
    m_app.assert_awaited_once_with({"type": "lifespan"}, m_receive, m_send)
    m_response_callback.assert_not_awaited()



from mrok.controller.auth.backends import OIDCJWTAuthenticationBackend  # noqa: F401
from mrok.controller.auth.base import AuthIdentity, BaseHTTPAuthBackend
from mrok.controller.auth.manager import HTTPAuthManager
from mrok.controller.auth.registry import register_authentication_backend

__all__ = [
    "AuthIdentity",
    "BaseHTTPAuthBackend",
    "HTTPAuthManager",
    "register_authentication_backend",
]
